<footer style="text-align: left; padding: 5px 20px; background: linear-gradient(225deg, #343a40, #495057); color: #ced4da; font-size: 12px;">
    <p style="margin: 0;">
        Logo by <a href="https://www.instagram.com/a_armin_art/profilecard/?igsh=Zno3c2Z4YzY5aXU4" target="_blank" style="color: #17a2b8; text-decoration: none;">Armin</a>
    </p>
  </footer>
  